/**
 * Copyright 2005 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.nutch.searcher.correction;

import org.apache.lucene.search.BooleanQuery;
import org.apache.lucene.search.PhraseQuery;
import org.apache.lucene.search.TermQuery;
import org.apache.lucene.search.FuzzyQuery;

import org.apache.nutch.analysis.NutchDocumentAnalyzer;
import org.apache.nutch.analysis.CommonGrams;

import org.apache.nutch.searcher.QueryFilter;
import org.apache.nutch.searcher.Query;
import org.apache.nutch.searcher.Query.*;
import org.apache.nutch.util.NutchConf;

import java.io.IOException;
import java.util.HashSet;

/** The correction query filter.  Query terms in the default query field are
 * expanded to search the url, anchor and content document fields.
 * This plugin should replace query-basic. It includes Lucene FuzzyQuery
 * features.
*/
public class CorrectionQueryFilter implements QueryFilter {
    
  private static float URL_BOOST =
    NutchConf.get().getFloat("query.url.boost", 4.0f);

  private static float ANCHOR_BOOST =
    NutchConf.get().getFloat("query.anchor.boost", 2.0f);

  private static float TITLE_BOOST =
    NutchConf.get().getFloat("query.title.boost", 1.5f);

    private static float HOST_BOOST =
      NutchConf.get().getFloat("query.host.boost", 2.0f);

  private static int SLOP = Integer.MAX_VALUE;

  private static float PHRASE_BOOST =
    NutchConf.get().getFloat("query.phrase.boost", 1.0f);

  private static final String[] FIELDS =
  { "url", "anchor", "content", "title", "host" };

  private static final float[] FIELD_BOOSTS =
  { URL_BOOST, ANCHOR_BOOST, 1.0f, TITLE_BOOST, HOST_BOOST };

  /**
   * Set the boost factor for url matches, relative to content and anchor
   * matches
   */
  public static void setUrlBoost(float boost) { URL_BOOST = boost; }

  /** Set the boost factor for title/anchor matches, relative to url and
   * content matches. */
  public static void setAnchorBoost(float boost) { ANCHOR_BOOST = boost; }

  /** Set the boost factor for sloppy phrase matches relative to unordered term
   * matches. */
  public static void setPhraseBoost(float boost) { PHRASE_BOOST = boost; }

  /** Set the maximum number of terms permitted between matching terms in a
   * sloppy phrase match. */
  public static void setSlop(int slop) { SLOP = slop; }

  public BooleanQuery filter(Query input, BooleanQuery output) {
    addTerms(input, output);
    addSloppyPhrases(input, output);
    return output;
  }

  private static void addTerms(Query input, BooleanQuery output) {
    Clause[] clauses = input.getClauses();
    boolean autocorrect = false;
    if(firstTerm.startsWith("autocorrectionmod"))
	{
	    autocorrect = true;
	}
    
    String firstTerm = clauses[0].getTerm().toString();
    for (int i = 0; i < clauses.length; i++) {
	if(i==0 && autocorrect )
	    {
		i++;
	    }
	else
	    {
		Clause c = clauses[i];
		if (!c.getField().equals(Clause.DEFAULT_FIELD))
		    continue;                                 // skip non-default fields
		
		BooleanQuery out = new BooleanQuery();
		for (int f = 0; f < FIELDS.length; f++) {
		    Clause o = c;
		    if (c.isPhrase()) {                         // optimize phrase clauses
			String[] opt = CommonGrams.optimizePhrase(c.getPhrase(), FIELDS[f]);
			if (opt.length==1) {
			    o = new Clause(new Term(opt[0]), c.isRequired(), c.isProhibited());
			} else {
			    o = new Clause(new Phrase(opt), c.isRequired(), c.isProhibited());
			}
		    }
		    
		     if (o.isPhrase()) {
			 out.add(exactPhrase(o.getPhrase(), FIELDS[f], FIELD_BOOSTS[f]));
		     }
		     else {
			 out.add(termQuery(FIELDS[f], o.getTerm(), FIELD_BOOSTS[f]), false, false);
			 // fuzzy query is really only appropriate for content, 
			 // and also its a bit slow, so don't do fuzzy anchors or urls;
			 // only do fuzzy for words 5 characters or more
			 if (autocorrect && "content".equals(FIELDS[f]) && (o.getTerm().toString().length() >= 5))
			     {
				 out.add(fuzzyQuery(FIELDS[f], o.getTerm()), false, false);
			     }
		     }
		}
		output.add(out, c.isRequired(), c.isProhibited());
	    }
    }
  }

     /** 
      * Utility to construct a Lucene fuzzy query. 
      */  
    private static org.apache.lucene.search.FuzzyQuery fuzzyQuery(String
								  field, Term term) {
	/**
	 * 0.65f means they have to get 65% of right characters for the (term.length - 2) 
	 * last characters.
	 * Used algorithm is Levenstein one.
	 */
	FuzzyQuery result = new FuzzyQuery(luceneTerm(field, term), 0.65f,2);
	// give fuzzy matches less boost in the overall search result ranking
	result.setBoost(0.5f);
	return result;
    }
    
    private static void addSloppyPhrases(Query input, BooleanQuery output) {
    Clause[] clauses = input.getClauses();
    for (int f = 0; f < FIELDS.length; f++) {

      PhraseQuery sloppyPhrase = new PhraseQuery();
      sloppyPhrase.setBoost(FIELD_BOOSTS[f] * PHRASE_BOOST);
      sloppyPhrase.setSlop("anchor".equals(FIELDS[f])
                           ? NutchDocumentAnalyzer.INTER_ANCHOR_GAP
                           : SLOP);
      int sloppyTerms = 0;

      for (int i = 0; i < clauses.length; i++) {
        Clause c = clauses[i];
        
        if (!c.getField().equals(Clause.DEFAULT_FIELD))
          continue;                               // skip non-default fields
        
        if (c.isPhrase())                         // skip exact phrases
          continue;

        if (c.isProhibited())                     // skip prohibited terms
          continue;
        
        sloppyPhrase.add(luceneTerm(FIELDS[f], c.getTerm()));
        sloppyTerms++;
      }

      if (sloppyTerms > 1)
        output.add(sloppyPhrase, false, false);
    }
  }


  private static org.apache.lucene.search.Query
        termQuery(String field, Term term, float boost) {
    TermQuery result = new TermQuery(luceneTerm(field, term));
    result.setBoost(boost);
    return result;
  }

  /** Utility to construct a Lucene exact phrase query for a Nutch phrase. */
  private static org.apache.lucene.search.Query
       exactPhrase(Phrase nutchPhrase,
                   String field, float boost) {
    Term[] terms = nutchPhrase.getTerms();
    PhraseQuery exactPhrase = new PhraseQuery();
    for (int i = 0; i < terms.length; i++) {
      exactPhrase.add(luceneTerm(field, terms[i]));
    }
    exactPhrase.setBoost(boost);
    return exactPhrase;
  }

  /** Utility to construct a Lucene Term given a Nutch query term and field. */
  private static org.apache.lucene.index.Term luceneTerm(String field,
                                                         Term term) {
    return new org.apache.lucene.index.Term(field, term.toString());
  }
}
